﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Media3D;
using System.Windows.Shapes;

namespace Matrix3DProjectionDemo
{
    public partial class MainPage : UserControl
    {
        Dot draggedDot;
        Point startPointMouse;
        Point startPointDot;
        Matrix3DProjection projection;

        public MainPage()
        {
            InitializeComponent();
            bigButtonRadioButton.IsChecked = true;
        }

        void RadioButton_Checked(object sender, RoutedEventArgs args)
        {
            UIElement child = null;

            elementHost.Children.Clear();

            if (sender == bigButtonRadioButton)
            {
                Button btn = new Button();
                btn.Content = "Button";
                btn.FontSize = 96;
                child = btn;
            }
            else if (sender == polygonsRadioButton)
            {
                child = new ClipArtFellow();
            }
            else if (sender == bitmapRadioButton)
            {
                Image img = new Image();
                img.Stretch = Stretch.None;
                Uri uri = new Uri("http://www.charlespetzold.com/PetzoldTattoo.jpg");
                BitmapImage bitmap = new BitmapImage(uri);
                bitmap.ImageOpened += OnBitmapImageOpened;
                img.Source = bitmap;
                child = img;
            }

            if (child != null)
            {
                projection = new Matrix3DProjection();
                child.Projection = projection;

                if (!(child is Image))
                    LayoutUpdated += OnLayoutUpdated;

                elementHost.Children.Add(child);
            }
        }

        void OnBitmapImageOpened(object sender, RoutedEventArgs e)
        {
            BitmapImage bitmap = sender as BitmapImage;
            bitmap.ImageOpened -= OnBitmapImageOpened;
            SetDots(new Size(bitmap.PixelWidth, bitmap.PixelHeight));
        }

        void OnLayoutUpdated(object sender, EventArgs args)
        {
            LayoutUpdated -= OnLayoutUpdated;
            SetDots(elementHost.RenderSize);
        }

        void SetDots(Size size)
        {
            dotUL.Center = new Point(0, 0);
            dotUR.Center = new Point(size.Width, 0);
            dotLL.Center = new Point(0, size.Height);
            dotLR.Center = new Point(size.Width, size.Height);
        }

        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs args)
        {
            base.OnMouseLeftButtonDown(args);

            if (args.OriginalSource is Path)
            {
                DependencyObject parent = (args.OriginalSource as Path).Parent;

                if (parent is Dot)
                {
                    draggedDot = parent as Dot;
                    startPointDot = draggedDot.Center;
                    startPointMouse = args.GetPosition(this);
                    CaptureMouse();
                }
            }
        }

        protected override void OnMouseMove(MouseEventArgs args)
        {
            base.OnMouseMove(args);

            if (draggedDot != null)
            {
                Point pointMouse = args.GetPosition(this);

                draggedDot.Center = new Point(startPointDot.X - startPointMouse.X + pointMouse.X,
                                              startPointDot.Y - startPointMouse.Y + pointMouse.Y);

                projection.ProjectionMatrix =
                matrix3Ddisplay.Matrix3D =
                    CalculateNewTransform(elementHost.RenderSize, 
                                          dotUL.Center, dotUR.Center, dotLL.Center, dotLR.Center);
            }
        }

        protected override void OnMouseLeftButtonUp(MouseButtonEventArgs args)
        {
            base.OnMouseLeftButtonUp(args);

            if (draggedDot != null)
            {
                ReleaseMouseCapture();
                draggedDot = null;
            }
        }

        Matrix3D CalculateNewTransform(Size size, Point ptUL, Point ptUR, Point ptLL, Point ptLR)
        {
            // Scale transform
            Matrix3D S = new Matrix3D();
            S.M11 = 1 / size.Width;
            S.M22 = 1 / size.Height;

            // See http://www.charlespetzold.com/blog/2007/08/250638.html for math that follows

            // Affine transform
            Matrix3D A = new Matrix3D();
            A.OffsetX = ptUL.X;
            A.OffsetY = ptUL.Y;
            A.M11 = (ptUR.X - A.OffsetX) / 1;
            A.M12 = (ptUR.Y - A.OffsetY) / 1;
            A.M21 = (ptLL.X - A.OffsetX) / 1;
            A.M22 = (ptLL.Y - A.OffsetY) / 1;

            // Non-Affine transform
            Matrix3D B = new Matrix3D();
            double den = A.M11 * A.M22 - A.M12 * A.M21;
            double a = (A.M22 * ptLR.X - A.M21 * ptLR.Y + A.M21 * A.OffsetY - A.M22 * A.OffsetX) / den;
            double b = (A.M11 * ptLR.Y - A.M12 * ptLR.X + A.M12 * A.OffsetX - A.M11 * A.OffsetY) / den;

            B.M11 = a / (a + b - 1);
            B.M22 = b / (a + b - 1);
            B.M14 = B.M11 - 1;
            B.M24 = B.M22 - 1;

            // Product of three transforms
            return S * B * A;
        }
    }
}
